import datetime
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.edit_action import EditAction
from ..models.split import Split
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.edit_create_response_edit_spans_type_0 import EditCreateResponseEditSpansType0
    from ..models.filter_params import FilterParams


T = TypeVar("T", bound="EditCreateResponse")


@_attrs_define
class EditCreateResponse:
    """
    Attributes:
        created_at (datetime.datetime):
        edit_action (EditAction): The available actions you can take in an edit.
        exported (bool):
        id (str):
        project_id (str):
        run_id (str):
        split (Split):
        created_by (Union[None, Unset, str]):
        edit_spans (Union['EditCreateResponseEditSpansType0', None, Unset]):
        filter_ (Union['FilterParams', None, Unset]):
        inference_name (Union[None, Unset, str]):
        map_threshold (Union[None, Unset, float]):
        new_label (Union[None, Unset, str]):
        note (Union[None, Unset, str]):
        sample_ids (Union[Unset, list[int]]):
        search_string (Union[None, Unset, str]):
        shift_span_end_num_words (Union[None, Unset, int]):
        shift_span_start_num_words (Union[None, Unset, int]):
        task (Union[None, Unset, str]):
        text_replacement (Union[None, Unset, str]):
        use_regex (Union[None, Unset, bool]):  Default: False.
    """

    created_at: datetime.datetime
    edit_action: EditAction
    exported: bool
    id: str
    project_id: str
    run_id: str
    split: Split
    created_by: Union[None, Unset, str] = UNSET
    edit_spans: Union["EditCreateResponseEditSpansType0", None, Unset] = UNSET
    filter_: Union["FilterParams", None, Unset] = UNSET
    inference_name: Union[None, Unset, str] = UNSET
    map_threshold: Union[None, Unset, float] = UNSET
    new_label: Union[None, Unset, str] = UNSET
    note: Union[None, Unset, str] = UNSET
    sample_ids: Union[Unset, list[int]] = UNSET
    search_string: Union[None, Unset, str] = UNSET
    shift_span_end_num_words: Union[None, Unset, int] = UNSET
    shift_span_start_num_words: Union[None, Unset, int] = UNSET
    task: Union[None, Unset, str] = UNSET
    text_replacement: Union[None, Unset, str] = UNSET
    use_regex: Union[None, Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.edit_create_response_edit_spans_type_0 import EditCreateResponseEditSpansType0
        from ..models.filter_params import FilterParams

        created_at = self.created_at.isoformat()

        edit_action = self.edit_action.value

        exported = self.exported

        id = self.id

        project_id = self.project_id

        run_id = self.run_id

        split = self.split.value

        created_by: Union[None, Unset, str]
        if isinstance(self.created_by, Unset):
            created_by = UNSET
        else:
            created_by = self.created_by

        edit_spans: Union[None, Unset, dict[str, Any]]
        if isinstance(self.edit_spans, Unset):
            edit_spans = UNSET
        elif isinstance(self.edit_spans, EditCreateResponseEditSpansType0):
            edit_spans = self.edit_spans.to_dict()
        else:
            edit_spans = self.edit_spans

        filter_: Union[None, Unset, dict[str, Any]]
        if isinstance(self.filter_, Unset):
            filter_ = UNSET
        elif isinstance(self.filter_, FilterParams):
            filter_ = self.filter_.to_dict()
        else:
            filter_ = self.filter_

        inference_name: Union[None, Unset, str]
        if isinstance(self.inference_name, Unset):
            inference_name = UNSET
        else:
            inference_name = self.inference_name

        map_threshold: Union[None, Unset, float]
        if isinstance(self.map_threshold, Unset):
            map_threshold = UNSET
        else:
            map_threshold = self.map_threshold

        new_label: Union[None, Unset, str]
        if isinstance(self.new_label, Unset):
            new_label = UNSET
        else:
            new_label = self.new_label

        note: Union[None, Unset, str]
        if isinstance(self.note, Unset):
            note = UNSET
        else:
            note = self.note

        sample_ids: Union[Unset, list[int]] = UNSET
        if not isinstance(self.sample_ids, Unset):
            sample_ids = self.sample_ids

        search_string: Union[None, Unset, str]
        if isinstance(self.search_string, Unset):
            search_string = UNSET
        else:
            search_string = self.search_string

        shift_span_end_num_words: Union[None, Unset, int]
        if isinstance(self.shift_span_end_num_words, Unset):
            shift_span_end_num_words = UNSET
        else:
            shift_span_end_num_words = self.shift_span_end_num_words

        shift_span_start_num_words: Union[None, Unset, int]
        if isinstance(self.shift_span_start_num_words, Unset):
            shift_span_start_num_words = UNSET
        else:
            shift_span_start_num_words = self.shift_span_start_num_words

        task: Union[None, Unset, str]
        if isinstance(self.task, Unset):
            task = UNSET
        else:
            task = self.task

        text_replacement: Union[None, Unset, str]
        if isinstance(self.text_replacement, Unset):
            text_replacement = UNSET
        else:
            text_replacement = self.text_replacement

        use_regex: Union[None, Unset, bool]
        if isinstance(self.use_regex, Unset):
            use_regex = UNSET
        else:
            use_regex = self.use_regex

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "created_at": created_at,
                "edit_action": edit_action,
                "exported": exported,
                "id": id,
                "project_id": project_id,
                "run_id": run_id,
                "split": split,
            }
        )
        if created_by is not UNSET:
            field_dict["created_by"] = created_by
        if edit_spans is not UNSET:
            field_dict["edit_spans"] = edit_spans
        if filter_ is not UNSET:
            field_dict["filter"] = filter_
        if inference_name is not UNSET:
            field_dict["inference_name"] = inference_name
        if map_threshold is not UNSET:
            field_dict["map_threshold"] = map_threshold
        if new_label is not UNSET:
            field_dict["new_label"] = new_label
        if note is not UNSET:
            field_dict["note"] = note
        if sample_ids is not UNSET:
            field_dict["sample_ids"] = sample_ids
        if search_string is not UNSET:
            field_dict["search_string"] = search_string
        if shift_span_end_num_words is not UNSET:
            field_dict["shift_span_end_num_words"] = shift_span_end_num_words
        if shift_span_start_num_words is not UNSET:
            field_dict["shift_span_start_num_words"] = shift_span_start_num_words
        if task is not UNSET:
            field_dict["task"] = task
        if text_replacement is not UNSET:
            field_dict["text_replacement"] = text_replacement
        if use_regex is not UNSET:
            field_dict["use_regex"] = use_regex

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.edit_create_response_edit_spans_type_0 import EditCreateResponseEditSpansType0
        from ..models.filter_params import FilterParams

        d = src_dict.copy()
        created_at = isoparse(d.pop("created_at"))

        edit_action = EditAction(d.pop("edit_action"))

        exported = d.pop("exported")

        id = d.pop("id")

        project_id = d.pop("project_id")

        run_id = d.pop("run_id")

        split = Split(d.pop("split"))

        def _parse_created_by(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        created_by = _parse_created_by(d.pop("created_by", UNSET))

        def _parse_edit_spans(data: object) -> Union["EditCreateResponseEditSpansType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                edit_spans_type_0 = EditCreateResponseEditSpansType0.from_dict(data)

                return edit_spans_type_0
            except:  # noqa: E722
                pass
            return cast(Union["EditCreateResponseEditSpansType0", None, Unset], data)

        edit_spans = _parse_edit_spans(d.pop("edit_spans", UNSET))

        def _parse_filter_(data: object) -> Union["FilterParams", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                filter_type_0 = FilterParams.from_dict(data)

                return filter_type_0
            except:  # noqa: E722
                pass
            return cast(Union["FilterParams", None, Unset], data)

        filter_ = _parse_filter_(d.pop("filter", UNSET))

        def _parse_inference_name(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        inference_name = _parse_inference_name(d.pop("inference_name", UNSET))

        def _parse_map_threshold(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        map_threshold = _parse_map_threshold(d.pop("map_threshold", UNSET))

        def _parse_new_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        new_label = _parse_new_label(d.pop("new_label", UNSET))

        def _parse_note(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        note = _parse_note(d.pop("note", UNSET))

        sample_ids = cast(list[int], d.pop("sample_ids", UNSET))

        def _parse_search_string(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        search_string = _parse_search_string(d.pop("search_string", UNSET))

        def _parse_shift_span_end_num_words(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        shift_span_end_num_words = _parse_shift_span_end_num_words(d.pop("shift_span_end_num_words", UNSET))

        def _parse_shift_span_start_num_words(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        shift_span_start_num_words = _parse_shift_span_start_num_words(d.pop("shift_span_start_num_words", UNSET))

        def _parse_task(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        task = _parse_task(d.pop("task", UNSET))

        def _parse_text_replacement(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        text_replacement = _parse_text_replacement(d.pop("text_replacement", UNSET))

        def _parse_use_regex(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        use_regex = _parse_use_regex(d.pop("use_regex", UNSET))

        edit_create_response = cls(
            created_at=created_at,
            edit_action=edit_action,
            exported=exported,
            id=id,
            project_id=project_id,
            run_id=run_id,
            split=split,
            created_by=created_by,
            edit_spans=edit_spans,
            filter_=filter_,
            inference_name=inference_name,
            map_threshold=map_threshold,
            new_label=new_label,
            note=note,
            sample_ids=sample_ids,
            search_string=search_string,
            shift_span_end_num_words=shift_span_end_num_words,
            shift_span_start_num_words=shift_span_start_num_words,
            task=task,
            text_replacement=text_replacement,
            use_regex=use_regex,
        )

        edit_create_response.additional_properties = d
        return edit_create_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.split import Split

T = TypeVar("T", bound="GetSplitsResponse")


@_attrs_define
class GetSplitsResponse:
    """
    Attributes:
        inference_comparison_splits (list[Split]):
        project_id (str):
        run_id (str):
        splits (list[Split]):
    """

    inference_comparison_splits: list[Split]
    project_id: str
    run_id: str
    splits: list[Split]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        inference_comparison_splits = []
        for inference_comparison_splits_item_data in self.inference_comparison_splits:
            inference_comparison_splits_item = inference_comparison_splits_item_data.value
            inference_comparison_splits.append(inference_comparison_splits_item)

        project_id = self.project_id

        run_id = self.run_id

        splits = []
        for splits_item_data in self.splits:
            splits_item = splits_item_data.value
            splits.append(splits_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "inference_comparison_splits": inference_comparison_splits,
                "project_id": project_id,
                "run_id": run_id,
                "splits": splits,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        inference_comparison_splits = []
        _inference_comparison_splits = d.pop("inference_comparison_splits")
        for inference_comparison_splits_item_data in _inference_comparison_splits:
            inference_comparison_splits_item = Split(inference_comparison_splits_item_data)

            inference_comparison_splits.append(inference_comparison_splits_item)

        project_id = d.pop("project_id")

        run_id = d.pop("run_id")

        splits = []
        _splits = d.pop("splits")
        for splits_item_data in _splits:
            splits_item = Split(splits_item_data)

            splits.append(splits_item)

        get_splits_response = cls(
            inference_comparison_splits=inference_comparison_splits, project_id=project_id, run_id=run_id, splits=splits
        )

        get_splits_response.additional_properties = d
        return get_splits_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class GroupLabel(str, Enum):
    AGENT_QUALITY = "Agent Quality"
    CONFIGURATION = "Configuration"
    CUSTOM_METRICS = "Custom Metrics"
    DATASET = "Dataset"
    GENERIC_PARAMETERS = "Generic Parameters"
    INPUT_QUALITY = "Input Quality"
    LLM_PARAMETERS = "LLM Parameters"
    OUTPUT_QUALITY = "Output Quality"
    PROTECT = "Protect"
    RAG_PARAMETERS = "RAG Parameters"
    RAG_QUALITY = "RAG Quality"
    RATING_SUMMARY = "Rating Summary"
    SAFETY_METRICS = "Safety Metrics"
    SECURITY = "Security"
    SYSTEM_METRICS = "System Metrics"

    def __str__(self) -> str:
        return str(self.value)



import contextvars
import json
import logging
import time
from typing import Any, Literal, Optional
from uuid import UUID

from galileo import galileo_context
from galileo.logger import GalileoLogger
from galileo.utils.serialization import EventSerializer, convert_to_string_dict, serialize_to_str

_logger = logging.getLogger("galileo.handlers.langchain")

try:
    from langchain_core.agents import AgentAction, AgentFinish
    from langchain_core.callbacks.base import BaseCallbackHandler
    from langchain_core.documents import Document
    from langchain_core.messages import BaseMessage
    from langchain_core.outputs import LLMResult
except ImportError:
    _logger.warning("Failed to import langchain, using stubs")
    BaseCallbackHandler = object
    Document = object
    AgentAction = object
    BaseMessage = object
    LLMResult = object
    AgentFinish = object


SPAN_TYPE = Literal["llm", "retriever", "tool", "workflow"]
LANGCHAIN_NODE_TYPE = Literal["agent", "chain", "chat", "llm", "retriever", "tool"]


class Node:
    """
    A node in the Langchain trace.

    Attributes
    ----------
    node_type : LANGCHAIN_NODE_TYPE
        The type of node.
    span_params : dict[str, Any]
        The parameters for the span that will be created.
    run_id : UUID
        The run ID of the span.
    parent_run_id : Optional[UUID]
        The run ID of the parent span.
    children : List[str]
        List of run_ids for child nodes
    """

    def __init__(
        self,
        node_type: LANGCHAIN_NODE_TYPE,
        span_params: dict[str, Any],
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
    ) -> None:
        self.node_type: LANGCHAIN_NODE_TYPE = node_type
        self.span_params: dict[str, Any] = span_params
        self.run_id: UUID = run_id
        self.parent_run_id: Optional[UUID] = parent_run_id
        self.children: list[str] = []


_root_node: contextvars.ContextVar[Optional[Node]] = contextvars.ContextVar("langchain_root_node", default=None)


class GalileoCallback(BaseCallbackHandler):
    """
    Langchain callback handler for logging traces to the Galileo platform.

    Attributes
    ----------
    _galileo_logger : GalileoLogger
        The Galileo logger instance.
    _nodes : dict[UUID, Node]
        A dictionary of nodes, where the key is the run_id and the value is the node.
    _start_new_trace : bool
        Whether to start a new trace when a chain starts. Set this to `False` to continue using the current trace.
    _flush_on_chain_end : bool
        Whether to flush the trace when a chain ends.
    """

    def __init__(
        self,
        galileo_logger: Optional[GalileoLogger] = None,
        start_new_trace: bool = True,
        flush_on_chain_end: bool = True,
    ):
        self._galileo_logger: GalileoLogger = galileo_logger or galileo_context.get_logger_instance()
        self._start_new_trace: bool = start_new_trace
        self._flush_on_chain_end: bool = flush_on_chain_end
        self._nodes: dict[str, Node] = {}

    def _commit(self) -> None:
        """
        Commit the nodes to the trace using the Galileo Logger. Optionally flush the trace.
        """
        if not self._nodes:
            _logger.warning("No nodes to commit")
            return

        root = _root_node.get()
        if root is None:
            _logger.warning("Unable to add nodes to trace: Root node not set")
            return

        root_node = self._nodes.get(str(root.run_id))
        if root_node is None:
            _logger.warning("Unable to add nodes to trace: Root node does not exist")
            return

        if self._start_new_trace:
            self._galileo_logger.start_trace(input=serialize_to_str(root_node.span_params.get("input", "")))

        self._log_node_tree(root_node)

        # Conclude the trace with the root node's output
        root_output = root_node.span_params.get("output", "")

        if self._start_new_trace:
            # If we started a new trace, we need to conclude it
            self._galileo_logger.conclude(output=serialize_to_str(root_output))

        if self._flush_on_chain_end:
            # Upload the trace to Galileo
            self._galileo_logger.flush()

        # Clear nodes after successful commit
        self._nodes.clear()
        _root_node.set(None)

    def _log_node_tree(self, node: Node) -> None:
        """
        Log a node and its children recursively.

        Parameters
        ----------
        node : Node
            The node to log.
        """
        is_workflow_span = False
        input = node.span_params.get("input", "")
        output = node.span_params.get("output", "")
        name = node.span_params.get("name")
        metadata = node.span_params.get("metadata")
        tags = node.span_params.get("tags")

        # Convert metadata to a dict[str, str]
        if metadata is not None:
            metadata = convert_to_string_dict(metadata)

        # Log the current node based on its type
        if node.node_type in ("agent", "chain"):
            self._galileo_logger.add_workflow_span(input=input, output=output, name=name, metadata=metadata, tags=tags)
            is_workflow_span = True
        elif node.node_type in ("llm", "chat"):
            self._galileo_logger.add_llm_span(
                input=input,
                output=output,
                model=node.span_params.get("model"),
                temperature=node.span_params.get("temperature"),
                name=name,
                metadata=metadata,
                tags=tags,
                num_input_tokens=node.span_params.get("num_input_tokens"),
                num_output_tokens=node.span_params.get("num_output_tokens"),
                total_tokens=node.span_params.get("total_tokens"),
                time_to_first_token_ns=node.span_params.get("time_to_first_token_ns"),
            )
        elif node.node_type == "retriever":
            self._galileo_logger.add_retriever_span(input=input, output=output, name=name, metadata=metadata, tags=tags)
        elif node.node_type == "tool":
            self._galileo_logger.add_tool_span(input=input, output=output, name=name, metadata=metadata, tags=tags)
        else:
            _logger.warning(f"Unknown node type: {node.node_type}")

        # Process all child nodes
        last_child = None
        for child_id in node.children:
            child_node = self._nodes.get(child_id)
            if child_node:
                self._log_node_tree(child_node)
                last_child = child_node
            else:
                _logger.warning(f"Child node {child_id} not found")

        # Conclude workflow span. Use the last child's output if necessary
        if is_workflow_span:
            output = output or (last_child.span_params.get("output", "") if last_child else "")
            self._galileo_logger.conclude(output=serialize_to_str(output))

    def _start_node(
        self, node_type: LANGCHAIN_NODE_TYPE, parent_run_id: Optional[UUID], run_id: UUID, **kwargs: Any
    ) -> Node:
        """
        Start a new node in the chain.

        Parameters
        ----------
        node_type : LANGCHAIN_NODE_TYPE
            The type of node.
        parent_run_id : Optional[UUID]
            The parent run ID.
        run_id : UUID
            The run ID.
        **kwargs : Any
            Additional parameters for the span.

        Returns
        -------
        Node
            The created node.
        """
        node_id = str(run_id)
        parent_node_id = str(parent_run_id) if parent_run_id else None

        if node_id in self._nodes:
            _logger.debug(f"Node already exists for run_id {run_id}, overwriting...")

        # Create new node
        node = Node(node_type=node_type, span_params=kwargs, run_id=run_id, parent_run_id=parent_run_id)
        self._nodes[node_id] = node

        # Set as root node if needed
        if not _root_node.get():
            _logger.debug(f"Setting root node to {node_id}")
            _root_node.set(node)

        # Add to parent's children if parent exists
        if parent_run_id:
            parent = self._nodes.get(parent_node_id)
            if parent:
                parent.children.append(node_id)
            else:
                _logger.debug(f"Parent node {parent_node_id} not found for {node_id}")

        return node

    def _end_node(self, run_id: UUID, **kwargs: Any) -> None:
        """
        End a node in the chain. Commit the nodes to a trace if the run_id matches the root node.

        Parameters
        ----------
        run_id : UUID
            The run ID.
        **kwargs : Any
            Additional parameters to update the span with.
        """
        node_id = str(run_id)
        node = self._nodes.get(node_id)

        if not node:
            _logger.debug(f"No node exists for run_id {node_id}")
            return

        # Update node parameters
        node.span_params.update(**kwargs)

        # Check if this is the root node and commit if so
        root = _root_node.get()
        if root and node.run_id == root.run_id:
            self._commit()

    def on_chain_start(
        self,
        serialized: dict[str, Any],
        inputs: dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> Any:
        """Langchain callback when a chain starts."""
        node_type = "chain"
        node_name = serialized.get("name", None) or kwargs.get("name", "Chain")

        # If the `name` is `LangGraph` or `agent`, set the node type to `agent`.
        if node_name in ["LangGraph", "agent"]:
            node_type = "agent"
            node_name = "Agent"

        kwargs["name"] = node_name

        # If the node is tagged with `hidden`, don't log it.
        if tags and "langsmith:hidden" in tags:
            return

        self._start_node(node_type, parent_run_id, run_id, input=serialize_to_str(inputs), tags=tags, **kwargs)

    def on_chain_end(
        self, outputs: dict[str, Any], *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any
    ) -> Any:
        """Langchain callback when a chain ends."""
        self._end_node(run_id, output=serialize_to_str(outputs))

    def on_agent_finish(self, finish: AgentFinish, *, run_id: UUID, **kwargs: Any) -> Any:
        """Langchain callback when an agent finishes."""
        self._end_node(run_id, output=serialize_to_str(finish))

    def on_llm_start(
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Langchain callback when an LLM node starts.

        Note: This callback is only used for non-chat models.
        """
        invocation_params = kwargs.get("invocation_params", {})
        model = invocation_params.get("model_name", "")
        temperature = invocation_params.get("temperature", 0.0)

        self._start_node(
            "llm",
            parent_run_id,
            run_id,
            name="LLM",
            input=[{"content": p, "role": "user"} for p in prompts],
            tags=tags,
            model=model,
            temperature=temperature,
            metadata={k: str(v) for k, v in metadata.items()} if metadata else None,
            start_time=time.perf_counter(),
            time_to_first_token_ns=None,
        )

    def on_llm_new_token(self, token: str, *, run_id: UUID, **kwargs: Any) -> Any:
        """Langchain callback when an LLM node generates a new token."""
        node = self._nodes.get(run_id)
        if not node:
            return

        if "time_to_first_token_ns" not in node.span_params or node.span_params["time_to_first_token_ns"] is None:
            start_time = node.span_params.get("start_time")
            if start_time is not None:
                node.span_params["time_to_first_token_ns"] = (time.perf_counter() - start_time) * 1e9

    def on_chat_model_start(
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Langchain callback when a chat model starts."""
        invocation_params = kwargs.get("invocation_params", {})
        model = invocation_params.get("model", invocation_params.get("_type", "undefined-type"))
        temperature = invocation_params.get("temperature", 0.0)

        # Serialize messages safely
        try:
            flattened_messages = [message.dict() for batch in messages for message in batch]
            serialized_messages = json.loads(json.dumps(flattened_messages, cls=EventSerializer))
        except Exception as e:
            _logger.warning(f"Failed to serialize chat messages: {e}")
            serialized_messages = str(messages)

        self._start_node(
            "chat",
            parent_run_id,
            run_id,
            name="Chat Model",
            input=serialized_messages,
            tags=tags,
            model=model,
            temperature=temperature,
            metadata={k: str(v) for k, v in metadata.items()} if metadata else None,
            time_to_first_token_ns=None,
        )

    def on_llm_end(
        self, response: LLMResult, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any
    ) -> Any:
        """Langchain callback when an LLM node ends."""
        token_usage = response.llm_output.get("token_usage", {}) if response.llm_output else {}

        try:
            flattened_messages = [message.dict() for batch in response.generations for message in batch]
            output = json.loads(json.dumps(flattened_messages[0], cls=EventSerializer))
        except Exception as e:
            _logger.warning(f"Failed to serialize LLM output: {e}")
            output = str(response.generations)

        self._end_node(
            run_id,
            output=output,
            num_input_tokens=token_usage.get("prompt_tokens"),
            num_output_tokens=token_usage.get("completion_tokens"),
            total_tokens=token_usage.get("total_tokens"),
        )

    def on_tool_start(
        self,
        serialized: dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Langchain callback when a tool node starts."""
        self._start_node(
            "tool",
            parent_run_id,
            run_id,
            name="Tool",
            input=input_str,
            tags=tags,
            metadata={k: str(v) for k, v in metadata.items()} if metadata else None,
        )

    def on_tool_end(self, output: str, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any) -> Any:
        """Langchain callback when a tool node ends."""
        self._end_node(run_id, output=serialize_to_str(output))

    def on_retriever_start(
        self, query: str, *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any
    ) -> Any:
        """Langchain callback when a retriever node starts."""
        self._start_node("retriever", parent_run_id, run_id, name="Retriever", input=serialize_to_str(query))

    def on_retriever_end(
        self, response: list[Document], *, run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any
    ) -> Any:
        """Langchain callback when a retriever node ends."""
        try:
            serialized_response = json.loads(json.dumps(response, cls=EventSerializer))
        except Exception as e:
            _logger.warning(f"Failed to serialize retriever output: {e}")
            serialized_response = str(response)

        self._end_node(run_id, output=serialized_response)



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.split import Split

if TYPE_CHECKING:
    from ..models.project_run_summary import ProjectRunSummary


T = TypeVar("T", bound="CompareRunsResponse")


@_attrs_define
class CompareRunsResponse:
    """
    Attributes:
        project_id (str):
        runs (list['ProjectRunSummary']):
        split (Split):
    """

    project_id: str
    runs: list["ProjectRunSummary"]
    split: Split
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        project_id = self.project_id

        runs = []
        for runs_item_data in self.runs:
            runs_item = runs_item_data.to_dict()
            runs.append(runs_item)

        split = self.split.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"project_id": project_id, "runs": runs, "split": split})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.project_run_summary import ProjectRunSummary

        d = src_dict.copy()
        project_id = d.pop("project_id")

        runs = []
        _runs = d.pop("runs")
        for runs_item_data in _runs:
            runs_item = ProjectRunSummary.from_dict(runs_item_data)

            runs.append(runs_item)

        split = Split(d.pop("split"))

        compare_runs_response = cls(project_id=project_id, runs=runs, split=split)

        compare_runs_response.additional_properties = d
        return compare_runs_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="MultiLabelTaskMetrics")


@_attrs_define
class MultiLabelTaskMetrics:
    """Metrics per task for multi-label models.

    Attributes:
        f1 (Union[Unset, list[float]]):
        labels (Union[Unset, list[Union[bool, int, str]]]):
        precision (Union[Unset, list[float]]):
        recall (Union[Unset, list[float]]):
    """

    f1: Union[Unset, list[float]] = UNSET
    labels: Union[Unset, list[Union[bool, int, str]]] = UNSET
    precision: Union[Unset, list[float]] = UNSET
    recall: Union[Unset, list[float]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        f1: Union[Unset, list[float]] = UNSET
        if not isinstance(self.f1, Unset):
            f1 = self.f1

        labels: Union[Unset, list[Union[bool, int, str]]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = []
            for labels_item_data in self.labels:
                labels_item: Union[bool, int, str]
                labels_item = labels_item_data
                labels.append(labels_item)

        precision: Union[Unset, list[float]] = UNSET
        if not isinstance(self.precision, Unset):
            precision = self.precision

        recall: Union[Unset, list[float]] = UNSET
        if not isinstance(self.recall, Unset):
            recall = self.recall

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if f1 is not UNSET:
            field_dict["f1"] = f1
        if labels is not UNSET:
            field_dict["labels"] = labels
        if precision is not UNSET:
            field_dict["precision"] = precision
        if recall is not UNSET:
            field_dict["recall"] = recall

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        f1 = cast(list[float], d.pop("f1", UNSET))

        labels = []
        _labels = d.pop("labels", UNSET)
        for labels_item_data in _labels or []:

            def _parse_labels_item(data: object) -> Union[bool, int, str]:
                return cast(Union[bool, int, str], data)

            labels_item = _parse_labels_item(labels_item_data)

            labels.append(labels_item)

        precision = cast(list[float], d.pop("precision", UNSET))

        recall = cast(list[float], d.pop("recall", UNSET))

        multi_label_task_metrics = cls(f1=f1, labels=labels, precision=precision, recall=recall)

        multi_label_task_metrics.additional_properties = d
        return multi_label_task_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateMetricCritiqueRequest")


@_attrs_define
class CreateMetricCritiqueRequest:
    """
    Attributes:
        critique (str):
        intended_value (bool): Set to opposite of actual value for new critiques, and set as same as existing critique
            intended value for updates.
        row_id (Union[UUID, int]):
        run_id (Union[None, str]):
        id (Union[None, Unset, str]): The id of the metric critique if it's being updated. None if it's a new critique.
    """

    critique: str
    intended_value: bool
    row_id: Union[UUID, int]
    run_id: Union[None, str]
    id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        critique = self.critique

        intended_value = self.intended_value

        row_id: Union[int, str]
        if isinstance(self.row_id, UUID):
            row_id = str(self.row_id)
        else:
            row_id = self.row_id

        run_id: Union[None, str]
        run_id = self.run_id

        id: Union[None, Unset, str]
        if isinstance(self.id, Unset):
            id = UNSET
        else:
            id = self.id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"critique": critique, "intended_value": intended_value, "row_id": row_id, "run_id": run_id})
        if id is not UNSET:
            field_dict["id"] = id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        critique = d.pop("critique")

        intended_value = d.pop("intended_value")

        def _parse_row_id(data: object) -> Union[UUID, int]:
            try:
                if not isinstance(data, str):
                    raise TypeError()
                row_id_type_0 = UUID(data)

                return row_id_type_0
            except:  # noqa: E722
                pass
            return cast(Union[UUID, int], data)

        row_id = _parse_row_id(d.pop("row_id"))

        def _parse_run_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        run_id = _parse_run_id(d.pop("run_id"))

        def _parse_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        id = _parse_id(d.pop("id", UNSET))

        create_metric_critique_request = cls(
            critique=critique, intended_value=intended_value, row_id=row_id, run_id=run_id, id=id
        )

        create_metric_critique_request.additional_properties = d
        return create_metric_critique_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class MessageRole(str, Enum):
    AGENT = "agent"
    ASSISTANT = "assistant"
    FUNCTION = "function"
    SYSTEM = "system"
    TOOL = "tool"
    USER = "user"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.b_box import BBox
    from ..models.edit_row_data_meta_type_0 import EditRowDataMetaType0
    from ..models.span import Span


T = TypeVar("T", bound="EditRowData")


@_attrs_define
class EditRowData:
    """
    Attributes:
        generated_output (Union[None, str]):
        id (int):
        input_ (Union[None, str]):
        target (Union[None, str]):
        boxes (Union[None, Unset, list['BBox']]):
        gold (Union[None, Unset, str]):
        image (Union[None, Unset, str]):  Default: ''.
        meta (Union['EditRowDataMetaType0', None, Unset]):
        pred (Union[None, Unset, str]):
        spans (Union[None, Unset, list['Span']]):
        text (Union[None, Unset, str]):
    """

    generated_output: Union[None, str]
    id: int
    input_: Union[None, str]
    target: Union[None, str]
    boxes: Union[None, Unset, list["BBox"]] = UNSET
    gold: Union[None, Unset, str] = UNSET
    image: Union[None, Unset, str] = ""
    meta: Union["EditRowDataMetaType0", None, Unset] = UNSET
    pred: Union[None, Unset, str] = UNSET
    spans: Union[None, Unset, list["Span"]] = UNSET
    text: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.edit_row_data_meta_type_0 import EditRowDataMetaType0

        generated_output: Union[None, str]
        generated_output = self.generated_output

        id = self.id

        input_: Union[None, str]
        input_ = self.input_

        target: Union[None, str]
        target = self.target

        boxes: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.boxes, Unset):
            boxes = UNSET
        elif isinstance(self.boxes, list):
            boxes = []
            for boxes_type_0_item_data in self.boxes:
                boxes_type_0_item = boxes_type_0_item_data.to_dict()
                boxes.append(boxes_type_0_item)

        else:
            boxes = self.boxes

        gold: Union[None, Unset, str]
        if isinstance(self.gold, Unset):
            gold = UNSET
        else:
            gold = self.gold

        image: Union[None, Unset, str]
        if isinstance(self.image, Unset):
            image = UNSET
        else:
            image = self.image

        meta: Union[None, Unset, dict[str, Any]]
        if isinstance(self.meta, Unset):
            meta = UNSET
        elif isinstance(self.meta, EditRowDataMetaType0):
            meta = self.meta.to_dict()
        else:
            meta = self.meta

        pred: Union[None, Unset, str]
        if isinstance(self.pred, Unset):
            pred = UNSET
        else:
            pred = self.pred

        spans: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.spans, Unset):
            spans = UNSET
        elif isinstance(self.spans, list):
            spans = []
            for spans_type_0_item_data in self.spans:
                spans_type_0_item = spans_type_0_item_data.to_dict()
                spans.append(spans_type_0_item)

        else:
            spans = self.spans

        text: Union[None, Unset, str]
        if isinstance(self.text, Unset):
            text = UNSET
        else:
            text = self.text

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"generated_output": generated_output, "id": id, "input": input_, "target": target})
        if boxes is not UNSET:
            field_dict["boxes"] = boxes
        if gold is not UNSET:
            field_dict["gold"] = gold
        if image is not UNSET:
            field_dict["image"] = image
        if meta is not UNSET:
            field_dict["meta"] = meta
        if pred is not UNSET:
            field_dict["pred"] = pred
        if spans is not UNSET:
            field_dict["spans"] = spans
        if text is not UNSET:
            field_dict["text"] = text

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.b_box import BBox
        from ..models.edit_row_data_meta_type_0 import EditRowDataMetaType0
        from ..models.span import Span

        d = src_dict.copy()

        def _parse_generated_output(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        generated_output = _parse_generated_output(d.pop("generated_output"))

        id = d.pop("id")

        def _parse_input_(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        input_ = _parse_input_(d.pop("input"))

        def _parse_target(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        target = _parse_target(d.pop("target"))

        def _parse_boxes(data: object) -> Union[None, Unset, list["BBox"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                boxes_type_0 = []
                _boxes_type_0 = data
                for boxes_type_0_item_data in _boxes_type_0:
                    boxes_type_0_item = BBox.from_dict(boxes_type_0_item_data)

                    boxes_type_0.append(boxes_type_0_item)

                return boxes_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["BBox"]], data)

        boxes = _parse_boxes(d.pop("boxes", UNSET))

        def _parse_gold(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        gold = _parse_gold(d.pop("gold", UNSET))

        def _parse_image(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        image = _parse_image(d.pop("image", UNSET))

        def _parse_meta(data: object) -> Union["EditRowDataMetaType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                meta_type_0 = EditRowDataMetaType0.from_dict(data)

                return meta_type_0
            except:  # noqa: E722
                pass
            return cast(Union["EditRowDataMetaType0", None, Unset], data)

        meta = _parse_meta(d.pop("meta", UNSET))

        def _parse_pred(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        pred = _parse_pred(d.pop("pred", UNSET))

        def _parse_spans(data: object) -> Union[None, Unset, list["Span"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                spans_type_0 = []
                _spans_type_0 = data
                for spans_type_0_item_data in _spans_type_0:
                    spans_type_0_item = Span.from_dict(spans_type_0_item_data)

                    spans_type_0.append(spans_type_0_item)

                return spans_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["Span"]], data)

        spans = _parse_spans(d.pop("spans", UNSET))

        def _parse_text(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        text = _parse_text(d.pop("text", UNSET))

        edit_row_data = cls(
            generated_output=generated_output,
            id=id,
            input_=input_,
            target=target,
            boxes=boxes,
            gold=gold,
            image=image,
            meta=meta,
            pred=pred,
            spans=spans,
            text=text,
        )

        edit_row_data.additional_properties = d
        return edit_row_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.chain_poll_template import ChainPollTemplate


T = TypeVar("T", bound="ScorerVersionResponse")


@_attrs_define
class ScorerVersionResponse:
    """
    Attributes:
        scorer_content (Union['ChainPollTemplate', str]): chainpoll template for llm based scorers and the code as a
            string for code based scorers
    """

    scorer_content: Union["ChainPollTemplate", str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.chain_poll_template import ChainPollTemplate

        scorer_content: Union[dict[str, Any], str]
        if isinstance(self.scorer_content, ChainPollTemplate):
            scorer_content = self.scorer_content.to_dict()
        else:
            scorer_content = self.scorer_content

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"scorer_content": scorer_content})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.chain_poll_template import ChainPollTemplate

        d = src_dict.copy()

        def _parse_scorer_content(data: object) -> Union["ChainPollTemplate", str]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                scorer_content_type_0 = ChainPollTemplate.from_dict(data)

                return scorer_content_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ChainPollTemplate", str], data)

        scorer_content = _parse_scorer_content(d.pop("scorer_content"))

        scorer_version_response = cls(scorer_content=scorer_content)

        scorer_version_response.additional_properties = d
        return scorer_version_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import builtins
import logging
from typing import Optional, Union

from galileo.base import BaseClientModel
from galileo.projects import Projects
from galileo.resources.api.prompts import (
    create_prompt_template_with_version_projects_project_id_templates_post,
    get_project_templates_projects_project_id_templates_get,
    get_template_from_project_projects_project_id_templates_template_id_get,
)
from galileo.resources.models import (
    BasePromptTemplateResponse,
    CreatePromptTemplateWithVersionRequestBody,
    HTTPValidationError,
    Message,
)

_logger = logging.getLogger(__name__)


class PromptTemplate(BasePromptTemplateResponse):
    def __init__(self, prompt_template: Union[None, BasePromptTemplateResponse] = None):
        """
        Initialize a PromptTemplate instance.

        Parameters
        ----------
        prompt_template : Union[None, BasePromptTemplateResponse], optional
            The prompt template data to initialize from. If None, creates an empty prompt template instance.
            Defaults to None.
        """
        if prompt_template is not None:
            super().__init__(
                all_available_versions=prompt_template.all_available_versions,
                id=prompt_template.id,
                max_version=prompt_template.max_version,
                name=prompt_template.name,
                selected_version=prompt_template.selected_version,
                selected_version_id=prompt_template.selected_version_id,
                template=prompt_template.template,
                total_versions=prompt_template.total_versions,
                all_versions=prompt_template.all_versions,
            )
            self.additional_properties = prompt_template.additional_properties.copy()
            return


class PromptTemplates(BaseClientModel):
    def list(self, project_name: str) -> list[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        templates = get_project_templates_projects_project_id_templates_get.sync(
            # TODO: remove type ignore, when migrated to proper AuthenticatedClient
            project_id=project.id,
            client=self.client,  # type: ignore[arg-type]
        )

        if not templates or isinstance(templates, HTTPValidationError):
            return []

        return [PromptTemplate(prompt_template=prompt_template) for prompt_template in templates]

    def get(self, *, project_name: str, template_id: str) -> Optional[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        _logger.debug(f"Get template {template_id} from project {project.id}")
        template = get_template_from_project_projects_project_id_templates_template_id_get.sync(
            # TODO: remove type ignore, when migrated to proper AuthenticatedClient
            project_id=project.id,
            template_id=template_id,
            client=self.client,  # type: ignore[arg-type]
        )

        if not template or isinstance(template, HTTPValidationError):
            return None

        return PromptTemplate(prompt_template=template)

    def create(
        self, name: str, project_name: str, template: Union[builtins.list[Message], str]
    ) -> Optional[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        body = CreatePromptTemplateWithVersionRequestBody(name=name, template=template)

        _logger.debug(f"{body}")
        response = create_prompt_template_with_version_projects_project_id_templates_post.sync_detailed(
            project_id=project.id,
            client=self.client,  # type: ignore[arg-type]
            body=body,
        )

        if not response.parsed or isinstance(response.parsed, HTTPValidationError):
            _logger.error(response)
            raise response

        return PromptTemplate(prompt_template=response.parsed)


def create_prompt_template(name: str, project: str, messages: Union[list["Message"], str]) -> Optional[PromptTemplate]:
    return PromptTemplates().create(name=name, project_name=project, template=messages)


def get_prompt_template(project: str, name: str) -> Optional[PromptTemplate]:
    prompt_templates = PromptTemplates().list(project_name=project)
    for prompt_template in prompt_templates:
        if prompt_template.name == name:
            _logger.info(f"Get template {prompt_template}")
            return prompt_template
    raise ValueError(f"Prompt {name} does not exist")



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.lasso_selection import LassoSelection
    from ..models.meta_filter import MetaFilter


T = TypeVar("T", bound="FilterParams")


@_attrs_define
class FilterParams:
    """
    Attributes:
        cbo_clusters (Union[None, Unset, list[int]]):
        class_filter (Union[None, Unset, list[str]]):
        cluster_ids (Union[None, Unset, list[int]]):
        confidence_high (Union[None, Unset, float]):
        confidence_low (Union[None, Unset, float]):
        correctly_classified (Union[None, Unset, bool]):
        data_embs (Union[None, Unset, bool]):  Default: False.
        data_error_potential_high (Union[None, Unset, float]):
        data_error_potential_low (Union[None, Unset, float]):
        drift_score_threshold (Union[None, Unset, float]):
        exclude_ids (Union[Unset, list[int]]):
        gold_filter (Union[None, Unset, list[str]]):
        ids (Union[Unset, list[int]]):
        image_ids (Union[None, Unset, list[int]]):
        is_drifted (Union[None, Unset, bool]):
        is_edited (Union[None, Unset, bool]):
        is_otb (Union[None, Unset, bool]):
        lasso (Union['LassoSelection', None, Unset]):
        likely_mislabeled (Union[None, Unset, bool]):
        likely_mislabeled_dep_percentile (Union[None, Unset, int]):  Default: 0.
        meta_filter (Union[None, Unset, list['MetaFilter']]):
        misclassified_only (Union[None, Unset, bool]):
        num_similar_to (Union[None, Unset, int]):
        pred_filter (Union[None, Unset, list[str]]):
        regex (Union[None, Unset, bool]):
        similar_to (Union[None, Unset, list[int]]):
        span_regex (Union[None, Unset, bool]):  Default: False.
        span_sample_ids (Union[None, Unset, list[int]]):
        span_text (Union[None, Unset, str]):
        text_pat (Union[None, Unset, str]):
    """

    cbo_clusters: Union[None, Unset, list[int]] = UNSET
    class_filter: Union[None, Unset, list[str]] = UNSET
    cluster_ids: Union[None, Unset, list[int]] = UNSET
    confidence_high: Union[None, Unset, float] = UNSET
    confidence_low: Union[None, Unset, float] = UNSET
    correctly_classified: Union[None, Unset, bool] = UNSET
    data_embs: Union[None, Unset, bool] = False
    data_error_potential_high: Union[None, Unset, float] = UNSET
    data_error_potential_low: Union[None, Unset, float] = UNSET
    drift_score_threshold: Union[None, Unset, float] = UNSET
    exclude_ids: Union[Unset, list[int]] = UNSET
    gold_filter: Union[None, Unset, list[str]] = UNSET
    ids: Union[Unset, list[int]] = UNSET
    image_ids: Union[None, Unset, list[int]] = UNSET
    is_drifted: Union[None, Unset, bool] = UNSET
    is_edited: Union[None, Unset, bool] = UNSET
    is_otb: Union[None, Unset, bool] = UNSET
    lasso: Union["LassoSelection", None, Unset] = UNSET
    likely_mislabeled: Union[None, Unset, bool] = UNSET
    likely_mislabeled_dep_percentile: Union[None, Unset, int] = 0
    meta_filter: Union[None, Unset, list["MetaFilter"]] = UNSET
    misclassified_only: Union[None, Unset, bool] = UNSET
    num_similar_to: Union[None, Unset, int] = UNSET
    pred_filter: Union[None, Unset, list[str]] = UNSET
    regex: Union[None, Unset, bool] = UNSET
    similar_to: Union[None, Unset, list[int]] = UNSET
    span_regex: Union[None, Unset, bool] = False
    span_sample_ids: Union[None, Unset, list[int]] = UNSET
    span_text: Union[None, Unset, str] = UNSET
    text_pat: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.lasso_selection import LassoSelection

        cbo_clusters: Union[None, Unset, list[int]]
        if isinstance(self.cbo_clusters, Unset):
            cbo_clusters = UNSET
        elif isinstance(self.cbo_clusters, list):
            cbo_clusters = self.cbo_clusters

        else:
            cbo_clusters = self.cbo_clusters

        class_filter: Union[None, Unset, list[str]]
        if isinstance(self.class_filter, Unset):
            class_filter = UNSET
        elif isinstance(self.class_filter, list):
            class_filter = self.class_filter

        else:
            class_filter = self.class_filter

        cluster_ids: Union[None, Unset, list[int]]
        if isinstance(self.cluster_ids, Unset):
            cluster_ids = UNSET
        elif isinstance(self.cluster_ids, list):
            cluster_ids = self.cluster_ids

        else:
            cluster_ids = self.cluster_ids

        confidence_high: Union[None, Unset, float]
        if isinstance(self.confidence_high, Unset):
            confidence_high = UNSET
        else:
            confidence_high = self.confidence_high

        confidence_low: Union[None, Unset, float]
        if isinstance(self.confidence_low, Unset):
            confidence_low = UNSET
        else:
            confidence_low = self.confidence_low

        correctly_classified: Union[None, Unset, bool]
        if isinstance(self.correctly_classified, Unset):
            correctly_classified = UNSET
        else:
            correctly_classified = self.correctly_classified

        data_embs: Union[None, Unset, bool]
        if isinstance(self.data_embs, Unset):
            data_embs = UNSET
        else:
            data_embs = self.data_embs

        data_error_potential_high: Union[None, Unset, float]
        if isinstance(self.data_error_potential_high, Unset):
            data_error_potential_high = UNSET
        else:
            data_error_potential_high = self.data_error_potential_high

        data_error_potential_low: Union[None, Unset, float]
        if isinstance(self.data_error_potential_low, Unset):
            data_error_potential_low = UNSET
        else:
            data_error_potential_low = self.data_error_potential_low

        drift_score_threshold: Union[None, Unset, float]
        if isinstance(self.drift_score_threshold, Unset):
            drift_score_threshold = UNSET
        else:
            drift_score_threshold = self.drift_score_threshold

        exclude_ids: Union[Unset, list[int]] = UNSET
        if not isinstance(self.exclude_ids, Unset):
            exclude_ids = self.exclude_ids

        gold_filter: Union[None, Unset, list[str]]
        if isinstance(self.gold_filter, Unset):
            gold_filter = UNSET
        elif isinstance(self.gold_filter, list):
            gold_filter = self.gold_filter

        else:
            gold_filter = self.gold_filter

        ids: Union[Unset, list[int]] = UNSET
        if not isinstance(self.ids, Unset):
            ids = self.ids

        image_ids: Union[None, Unset, list[int]]
        if isinstance(self.image_ids, Unset):
            image_ids = UNSET
        elif isinstance(self.image_ids, list):
            image_ids = self.image_ids

        else:
            image_ids = self.image_ids

        is_drifted: Union[None, Unset, bool]
        if isinstance(self.is_drifted, Unset):
            is_drifted = UNSET
        else:
            is_drifted = self.is_drifted

        is_edited: Union[None, Unset, bool]
        if isinstance(self.is_edited, Unset):
            is_edited = UNSET
        else:
            is_edited = self.is_edited

        is_otb: Union[None, Unset, bool]
        if isinstance(self.is_otb, Unset):
            is_otb = UNSET
        else:
            is_otb = self.is_otb

        lasso: Union[None, Unset, dict[str, Any]]
        if isinstance(self.lasso, Unset):
            lasso = UNSET
        elif isinstance(self.lasso, LassoSelection):
            lasso = self.lasso.to_dict()
        else:
            lasso = self.lasso

        likely_mislabeled: Union[None, Unset, bool]
        if isinstance(self.likely_mislabeled, Unset):
            likely_mislabeled = UNSET
        else:
            likely_mislabeled = self.likely_mislabeled

        likely_mislabeled_dep_percentile: Union[None, Unset, int]
        if isinstance(self.likely_mislabeled_dep_percentile, Unset):
            likely_mislabeled_dep_percentile = UNSET
        else:
            likely_mislabeled_dep_percentile = self.likely_mislabeled_dep_percentile

        meta_filter: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.meta_filter, Unset):
            meta_filter = UNSET
        elif isinstance(self.meta_filter, list):
            meta_filter = []
            for meta_filter_type_0_item_data in self.meta_filter:
                meta_filter_type_0_item = meta_filter_type_0_item_data.to_dict()
                meta_filter.append(meta_filter_type_0_item)

        else:
            meta_filter = self.meta_filter

        misclassified_only: Union[None, Unset, bool]
        if isinstance(self.misclassified_only, Unset):
            misclassified_only = UNSET
        else:
            misclassified_only = self.misclassified_only

        num_similar_to: Union[None, Unset, int]
        if isinstance(self.num_similar_to, Unset):
            num_similar_to = UNSET
        else:
            num_similar_to = self.num_similar_to

        pred_filter: Union[None, Unset, list[str]]
        if isinstance(self.pred_filter, Unset):
            pred_filter = UNSET
        elif isinstance(self.pred_filter, list):
            pred_filter = self.pred_filter

        else:
            pred_filter = self.pred_filter

        regex: Union[None, Unset, bool]
        if isinstance(self.regex, Unset):
            regex = UNSET
        else:
            regex = self.regex

        similar_to: Union[None, Unset, list[int]]
        if isinstance(self.similar_to, Unset):
            similar_to = UNSET
        elif isinstance(self.similar_to, list):
            similar_to = self.similar_to

        else:
            similar_to = self.similar_to

        span_regex: Union[None, Unset, bool]
        if isinstance(self.span_regex, Unset):
            span_regex = UNSET
        else:
            span_regex = self.span_regex

        span_sample_ids: Union[None, Unset, list[int]]
        if isinstance(self.span_sample_ids, Unset):
            span_sample_ids = UNSET
        elif isinstance(self.span_sample_ids, list):
            span_sample_ids = self.span_sample_ids

        else:
            span_sample_ids = self.span_sample_ids

        span_text: Union[None, Unset, str]
        if isinstance(self.span_text, Unset):
            span_text = UNSET
        else:
            span_text = self.span_text

        text_pat: Union[None, Unset, str]
        if isinstance(self.text_pat, Unset):
            text_pat = UNSET
        else:
            text_pat = self.text_pat

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if cbo_clusters is not UNSET:
            field_dict["cbo_clusters"] = cbo_clusters
        if class_filter is not UNSET:
            field_dict["class_filter"] = class_filter
        if cluster_ids is not UNSET:
            field_dict["cluster_ids"] = cluster_ids
        if confidence_high is not UNSET:
            field_dict["confidence_high"] = confidence_high
        if confidence_low is not UNSET:
            field_dict["confidence_low"] = confidence_low
        if correctly_classified is not UNSET:
            field_dict["correctly_classified"] = correctly_classified
        if data_embs is not UNSET:
            field_dict["data_embs"] = data_embs
        if data_error_potential_high is not UNSET:
            field_dict["data_error_potential_high"] = data_error_potential_high
        if data_error_potential_low is not UNSET:
            field_dict["data_error_potential_low"] = data_error_potential_low
        if drift_score_threshold is not UNSET:
            field_dict["drift_score_threshold"] = drift_score_threshold
        if exclude_ids is not UNSET:
            field_dict["exclude_ids"] = exclude_ids
        if gold_filter is not UNSET:
            field_dict["gold_filter"] = gold_filter
        if ids is not UNSET:
            field_dict["ids"] = ids
        if image_ids is not UNSET:
            field_dict["image_ids"] = image_ids
        if is_drifted is not UNSET:
            field_dict["is_drifted"] = is_drifted
        if is_edited is not UNSET:
            field_dict["is_edited"] = is_edited
        if is_otb is not UNSET:
            field_dict["is_otb"] = is_otb
        if lasso is not UNSET:
            field_dict["lasso"] = lasso
        if likely_mislabeled is not UNSET:
            field_dict["likely_mislabeled"] = likely_mislabeled
        if likely_mislabeled_dep_percentile is not UNSET:
            field_dict["likely_mislabeled_dep_percentile"] = likely_mislabeled_dep_percentile
        if meta_filter is not UNSET:
            field_dict["meta_filter"] = meta_filter
        if misclassified_only is not UNSET:
            field_dict["misclassified_only"] = misclassified_only
        if num_similar_to is not UNSET:
            field_dict["num_similar_to"] = num_similar_to
        if pred_filter is not UNSET:
            field_dict["pred_filter"] = pred_filter
        if regex is not UNSET:
            field_dict["regex"] = regex
        if similar_to is not UNSET:
            field_dict["similar_to"] = similar_to
        if span_regex is not UNSET:
            field_dict["span_regex"] = span_regex
        if span_sample_ids is not UNSET:
            field_dict["span_sample_ids"] = span_sample_ids
        if span_text is not UNSET:
            field_dict["span_text"] = span_text
        if text_pat is not UNSET:
            field_dict["text_pat"] = text_pat

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.lasso_selection import LassoSelection
        from ..models.meta_filter import MetaFilter

        d = src_dict.copy()

        def _parse_cbo_clusters(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                cbo_clusters_type_0 = cast(list[int], data)

                return cbo_clusters_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        cbo_clusters = _parse_cbo_clusters(d.pop("cbo_clusters", UNSET))

        def _parse_class_filter(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                class_filter_type_0 = cast(list[str], data)

                return class_filter_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        class_filter = _parse_class_filter(d.pop("class_filter", UNSET))

        def _parse_cluster_ids(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                cluster_ids_type_0 = cast(list[int], data)

                return cluster_ids_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        cluster_ids = _parse_cluster_ids(d.pop("cluster_ids", UNSET))

        def _parse_confidence_high(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        confidence_high = _parse_confidence_high(d.pop("confidence_high", UNSET))

        def _parse_confidence_low(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        confidence_low = _parse_confidence_low(d.pop("confidence_low", UNSET))

        def _parse_correctly_classified(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        correctly_classified = _parse_correctly_classified(d.pop("correctly_classified", UNSET))

        def _parse_data_embs(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        data_embs = _parse_data_embs(d.pop("data_embs", UNSET))

        def _parse_data_error_potential_high(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        data_error_potential_high = _parse_data_error_potential_high(d.pop("data_error_potential_high", UNSET))

        def _parse_data_error_potential_low(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        data_error_potential_low = _parse_data_error_potential_low(d.pop("data_error_potential_low", UNSET))

        def _parse_drift_score_threshold(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        drift_score_threshold = _parse_drift_score_threshold(d.pop("drift_score_threshold", UNSET))

        exclude_ids = cast(list[int], d.pop("exclude_ids", UNSET))

        def _parse_gold_filter(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                gold_filter_type_0 = cast(list[str], data)

                return gold_filter_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        gold_filter = _parse_gold_filter(d.pop("gold_filter", UNSET))

        ids = cast(list[int], d.pop("ids", UNSET))

        def _parse_image_ids(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                image_ids_type_0 = cast(list[int], data)

                return image_ids_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        image_ids = _parse_image_ids(d.pop("image_ids", UNSET))

        def _parse_is_drifted(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        is_drifted = _parse_is_drifted(d.pop("is_drifted", UNSET))

        def _parse_is_edited(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        is_edited = _parse_is_edited(d.pop("is_edited", UNSET))

        def _parse_is_otb(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        is_otb = _parse_is_otb(d.pop("is_otb", UNSET))

        def _parse_lasso(data: object) -> Union["LassoSelection", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                lasso_type_0 = LassoSelection.from_dict(data)

                return lasso_type_0
            except:  # noqa: E722
                pass
            return cast(Union["LassoSelection", None, Unset], data)

        lasso = _parse_lasso(d.pop("lasso", UNSET))

        def _parse_likely_mislabeled(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        likely_mislabeled = _parse_likely_mislabeled(d.pop("likely_mislabeled", UNSET))

        def _parse_likely_mislabeled_dep_percentile(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        likely_mislabeled_dep_percentile = _parse_likely_mislabeled_dep_percentile(
            d.pop("likely_mislabeled_dep_percentile", UNSET)
        )

        def _parse_meta_filter(data: object) -> Union[None, Unset, list["MetaFilter"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                meta_filter_type_0 = []
                _meta_filter_type_0 = data
                for meta_filter_type_0_item_data in _meta_filter_type_0:
                    meta_filter_type_0_item = MetaFilter.from_dict(meta_filter_type_0_item_data)

                    meta_filter_type_0.append(meta_filter_type_0_item)

                return meta_filter_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["MetaFilter"]], data)

        meta_filter = _parse_meta_filter(d.pop("meta_filter", UNSET))

        def _parse_misclassified_only(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        misclassified_only = _parse_misclassified_only(d.pop("misclassified_only", UNSET))

        def _parse_num_similar_to(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        num_similar_to = _parse_num_similar_to(d.pop("num_similar_to", UNSET))

        def _parse_pred_filter(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                pred_filter_type_0 = cast(list[str], data)

                return pred_filter_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        pred_filter = _parse_pred_filter(d.pop("pred_filter", UNSET))

        def _parse_regex(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        regex = _parse_regex(d.pop("regex", UNSET))

        def _parse_similar_to(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                similar_to_type_0 = cast(list[int], data)

                return similar_to_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        similar_to = _parse_similar_to(d.pop("similar_to", UNSET))

        def _parse_span_regex(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        span_regex = _parse_span_regex(d.pop("span_regex", UNSET))

        def _parse_span_sample_ids(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                span_sample_ids_type_0 = cast(list[int], data)

                return span_sample_ids_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        span_sample_ids = _parse_span_sample_ids(d.pop("span_sample_ids", UNSET))

        def _parse_span_text(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        span_text = _parse_span_text(d.pop("span_text", UNSET))

        def _parse_text_pat(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        text_pat = _parse_text_pat(d.pop("text_pat", UNSET))

        filter_params = cls(
            cbo_clusters=cbo_clusters,
            class_filter=class_filter,
            cluster_ids=cluster_ids,
            confidence_high=confidence_high,
            confidence_low=confidence_low,
            correctly_classified=correctly_classified,
            data_embs=data_embs,
            data_error_potential_high=data_error_potential_high,
            data_error_potential_low=data_error_potential_low,
            drift_score_threshold=drift_score_threshold,
            exclude_ids=exclude_ids,
            gold_filter=gold_filter,
            ids=ids,
            image_ids=image_ids,
            is_drifted=is_drifted,
            is_edited=is_edited,
            is_otb=is_otb,
            lasso=lasso,
            likely_mislabeled=likely_mislabeled,
            likely_mislabeled_dep_percentile=likely_mislabeled_dep_percentile,
            meta_filter=meta_filter,
            misclassified_only=misclassified_only,
            num_similar_to=num_similar_to,
            pred_filter=pred_filter,
            regex=regex,
            similar_to=similar_to,
            span_regex=span_regex,
            span_sample_ids=span_sample_ids,
            span_text=span_text,
            text_pat=text_pat,
        )

        filter_params.additional_properties = d
        return filter_params

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.transaction_logging_method import TransactionLoggingMethod
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.transaction_record_ingest import TransactionRecordIngest


T = TypeVar("T", bound="TransactionRecordBatch")


@_attrs_define
class TransactionRecordBatch:
    """
    Attributes:
        records (list['TransactionRecordIngest']):
        client_version (Union[None, Unset, str]):
        logging_method (Union[Unset, TransactionLoggingMethod]):
    """

    records: list["TransactionRecordIngest"]
    client_version: Union[None, Unset, str] = UNSET
    logging_method: Union[Unset, TransactionLoggingMethod] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        records = []
        for records_item_data in self.records:
            records_item = records_item_data.to_dict()
            records.append(records_item)

        client_version: Union[None, Unset, str]
        if isinstance(self.client_version, Unset):
            client_version = UNSET
        else:
            client_version = self.client_version

        logging_method: Union[Unset, str] = UNSET
        if not isinstance(self.logging_method, Unset):
            logging_method = self.logging_method.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"records": records})
        if client_version is not UNSET:
            field_dict["client_version"] = client_version
        if logging_method is not UNSET:
            field_dict["logging_method"] = logging_method

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.transaction_record_ingest import TransactionRecordIngest

        d = src_dict.copy()
        records = []
        _records = d.pop("records")
        for records_item_data in _records:
            records_item = TransactionRecordIngest.from_dict(records_item_data)

            records.append(records_item)

        def _parse_client_version(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        client_version = _parse_client_version(d.pop("client_version", UNSET))

        _logging_method = d.pop("logging_method", UNSET)
        logging_method: Union[Unset, TransactionLoggingMethod]
        if isinstance(_logging_method, Unset):
            logging_method = UNSET
        else:
            logging_method = TransactionLoggingMethod(_logging_method)

        transaction_record_batch = cls(records=records, client_version=client_version, logging_method=logging_method)

        transaction_record_batch.additional_properties = d
        return transaction_record_batch

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class LLMIntegration(str, Enum):
    ANTHROPIC = "anthropic"
    AWS_BEDROCK = "aws_bedrock"
    AWS_SAGEMAKER = "aws_sagemaker"
    AZURE = "azure"
    DATABRICKS = "databricks"
    MISTRAL = "mistral"
    OPENAI = "openai"
    VERTEX_AI = "vertex_ai"
    WRITER = "writer"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.group_by_metrics import GroupByMetrics


T = TypeVar("T", bound="Distribution")


@_attrs_define
class Distribution:
    """A distribution histogram for a run/filter for a particular numerical column.

    bins are boundaries for the x axis and counts are the y axis. bins will have 1 more value than counts because the
    count at index i is for the range [bin[i], bin[i+1]), with the lower boundary inclusive and the upper boundary
    exclusive

        Attributes:
            bins (list[float]):
            counts (list[int]):
            metrics (Union['GroupByMetrics', None, Unset]):
    """

    bins: list[float]
    counts: list[int]
    metrics: Union["GroupByMetrics", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.group_by_metrics import GroupByMetrics

        bins = self.bins

        counts = self.counts

        metrics: Union[None, Unset, dict[str, Any]]
        if isinstance(self.metrics, Unset):
            metrics = UNSET
        elif isinstance(self.metrics, GroupByMetrics):
            metrics = self.metrics.to_dict()
        else:
            metrics = self.metrics

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"bins": bins, "counts": counts})
        if metrics is not UNSET:
            field_dict["metrics"] = metrics

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.group_by_metrics import GroupByMetrics

        d = src_dict.copy()
        bins = cast(list[float], d.pop("bins"))

        counts = cast(list[int], d.pop("counts"))

        def _parse_metrics(data: object) -> Union["GroupByMetrics", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                metrics_type_0 = GroupByMetrics.from_dict(data)

                return metrics_type_0
            except:  # noqa: E722
                pass
            return cast(Union["GroupByMetrics", None, Unset], data)

        metrics = _parse_metrics(d.pop("metrics", UNSET))

        distribution = cls(bins=bins, counts=counts, metrics=metrics)

        distribution.additional_properties = d
        return distribution

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.data_row import DataRow


T = TypeVar("T", bound="DataRows")


@_attrs_define
class DataRows:
    """
    Attributes:
        data_rows (list['DataRow']):
        has_next_page (bool):
    """

    data_rows: list["DataRow"]
    has_next_page: bool
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        data_rows = []
        for data_rows_item_data in self.data_rows:
            data_rows_item = data_rows_item_data.to_dict()
            data_rows.append(data_rows_item)

        has_next_page = self.has_next_page

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"data_rows": data_rows, "has_next_page": has_next_page})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.data_row import DataRow

        d = src_dict.copy()
        data_rows = []
        _data_rows = d.pop("data_rows")
        for data_rows_item_data in _data_rows:
            data_rows_item = DataRow.from_dict(data_rows_item_data)

            data_rows.append(data_rows_item)

        has_next_page = d.pop("has_next_page")

        data_rows = cls(data_rows=data_rows, has_next_page=has_next_page)

        data_rows.additional_properties = d
        return data_rows

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class NodeType(str, Enum):
    AGENT = "agent"
    CHAIN = "chain"
    CHAT = "chat"
    LLM = "llm"
    RETRIEVER = "retriever"
    TOOL = "tool"
    TRACE = "trace"
    WORKFLOW = "workflow"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="IndexSelector")


@_attrs_define
class IndexSelector:
    """Choose specific indexes to apply the bulk operation to.

    Attributes:
        indexes (list[int]):
        selector_type (Union[Literal['indexes'], Unset]):  Default: 'indexes'.
    """

    indexes: list[int]
    selector_type: Union[Literal["indexes"], Unset] = "indexes"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        indexes = self.indexes

        selector_type = self.selector_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"indexes": indexes})
        if selector_type is not UNSET:
            field_dict["selector_type"] = selector_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        indexes = cast(list[int], d.pop("indexes"))

        selector_type = cast(Union[Literal["indexes"], Unset], d.pop("selector_type", UNSET))
        if selector_type != "indexes" and not isinstance(selector_type, Unset):
            raise ValueError(f"selector_type must match const 'indexes', got '{selector_type}'")

        index_selector = cls(indexes=indexes, selector_type=selector_type)

        index_selector.additional_properties = d
        return index_selector

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="BulkFeedbackRatingDelete")


@_attrs_define
class BulkFeedbackRatingDelete:
    """
    Attributes:
        feedback_template_id (str):
        operation_type (Union[Literal['delete'], Unset]):  Default: 'delete'.
    """

    feedback_template_id: str
    operation_type: Union[Literal["delete"], Unset] = "delete"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        feedback_template_id = self.feedback_template_id

        operation_type = self.operation_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"feedback_template_id": feedback_template_id})
        if operation_type is not UNSET:
            field_dict["operation_type"] = operation_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        feedback_template_id = d.pop("feedback_template_id")

        operation_type = cast(Union[Literal["delete"], Unset], d.pop("operation_type", UNSET))
        if operation_type != "delete" and not isinstance(operation_type, Unset):
            raise ValueError(f"operation_type must match const 'delete', got '{operation_type}'")

        bulk_feedback_rating_delete = cls(feedback_template_id=feedback_template_id, operation_type=operation_type)

        bulk_feedback_rating_delete.additional_properties = d
        return bulk_feedback_rating_delete

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.edit_content import EditContent
from ...models.http_validation_error import HTTPValidationError
from ...types import UNSET, Response, Unset


def _get_kwargs(
    edit_id: str,
    *,
    page_num: Union[Unset, int] = 0,
    meta_cols: Union[None, Unset, list[str]] = UNSET,
    text_pat: Union[None, Unset, str] = UNSET,
    regex: Union[None, Unset, bool] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["page_num"] = page_num

    json_meta_cols: Union[None, Unset, list[str]]
    if isinstance(meta_cols, Unset):
        json_meta_cols = UNSET
    elif isinstance(meta_cols, list):
        json_meta_cols = meta_cols

    else:
        json_meta_cols = meta_cols
    params["meta_cols"] = json_meta_cols

    json_text_pat: Union[None, Unset, str]
    if isinstance(text_pat, Unset):
        json_text_pat = UNSET
    else:
        json_text_pat = text_pat
    params["text_pat"] = json_text_pat

    json_regex: Union[None, Unset, bool]
    if isinstance(regex, Unset):
        json_regex = UNSET
    else:
        json_regex = regex
    params["regex"] = json_regex

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {"method": "get", "url": f"/edits/{edit_id}/content", "params": params}

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[EditContent, HTTPValidationError]]:
    if response.status_code == 200:
        response_200 = EditContent.from_dict(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[EditContent, HTTPValidationError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    edit_id: str,
    *,
    client: AuthenticatedClient,
    page_num: Union[Unset, int] = 0,
    meta_cols: Union[None, Unset, list[str]] = UNSET,
    text_pat: Union[None, Unset, str] = UNSET,
    regex: Union[None, Unset, bool] = UNSET,
) -> Response[Union[EditContent, HTTPValidationError]]:
    """Get Edit Content

     Gets the data for an edit, including the text, predicted label and true label.

    Args:
        edit_id (str):
        page_num (Union[Unset, int]):  Default: 0.
        meta_cols (Union[None, Unset, list[str]]):
        text_pat (Union[None, Unset, str]):
        regex (Union[None, Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[EditContent, HTTPValidationError]]
    """

    kwargs = _get_kwargs(edit_id=edit_id, page_num=page_num, meta_cols=meta_cols, text_pat=text_pat, regex=regex)

    response = client.get_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    edit_id: str,
    *,
    client: AuthenticatedClient,
    page_num: Union[Unset, int] = 0,
    meta_cols: Union[None, Unset, list[str]] = UNSET,
    text_pat: Union[None, Unset, str] = UNSET,
    regex: Union[None, Unset, bool] = UNSET,
) -> Optional[Union[EditContent, HTTPValidationError]]:
    """Get Edit Content

     Gets the data for an edit, including the text, predicted label and true label.

    Args:
        edit_id (str):
        page_num (Union[Unset, int]):  Default: 0.
        meta_cols (Union[None, Unset, list[str]]):
        text_pat (Union[None, Unset, str]):
        regex (Union[None, Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[EditContent, HTTPValidationError]
    """

    return sync_detailed(
        edit_id=edit_id, client=client, page_num=page_num, meta_cols=meta_cols, text_pat=text_pat, regex=regex
    ).parsed


async def asyncio_detailed(
    edit_id: str,
    *,
    client: AuthenticatedClient,
    page_num: Union[Unset, int] = 0,
    meta_cols: Union[None, Unset, list[str]] = UNSET,
    text_pat: Union[None, Unset, str] = UNSET,
    regex: Union[None, Unset, bool] = UNSET,
) -> Response[Union[EditContent, HTTPValidationError]]:
    """Get Edit Content

     Gets the data for an edit, including the text, predicted label and true label.

    Args:
        edit_id (str):
        page_num (Union[Unset, int]):  Default: 0.
        meta_cols (Union[None, Unset, list[str]]):
        text_pat (Union[None, Unset, str]):
        regex (Union[None, Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[EditContent, HTTPValidationError]]
    """

    kwargs = _get_kwargs(edit_id=edit_id, page_num=page_num, meta_cols=meta_cols, text_pat=text_pat, regex=regex)

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    edit_id: str,
    *,
    client: AuthenticatedClient,
    page_num: Union[Unset, int] = 0,
    meta_cols: Union[None, Unset, list[str]] = UNSET,
    text_pat: Union[None, Unset, str] = UNSET,
    regex: Union[None, Unset, bool] = UNSET,
) -> Optional[Union[EditContent, HTTPValidationError]]:
    """Get Edit Content

     Gets the data for an edit, including the text, predicted label and true label.

    Args:
        edit_id (str):
        page_num (Union[Unset, int]):  Default: 0.
        meta_cols (Union[None, Unset, list[str]]):
        text_pat (Union[None, Unset, str]):
        regex (Union[None, Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[EditContent, HTTPValidationError]
    """

    return (
        await asyncio_detailed(
            edit_id=edit_id, client=client, page_num=page_num, meta_cols=meta_cols, text_pat=text_pat, regex=regex
        )
    ).parsed



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.run_tag_create_request import RunTagCreateRequest


T = TypeVar("T", bound="RunCreate")


@_attrs_define
class RunCreate:
    """
    Attributes:
        name (str):
        task_type (int):
        created_by (Union[None, Unset, str]):
        dataset_hash (Union[None, Unset, str]):
        num_samples (Union[None, Unset, int]):
        project_id (Union[None, Unset, str]):
        run_tags (Union[Unset, list['RunTagCreateRequest']]):
        winner (Union[None, Unset, bool]):
    """

    name: str
    task_type: int
    created_by: Union[None, Unset, str] = UNSET
    dataset_hash: Union[None, Unset, str] = UNSET
    num_samples: Union[None, Unset, int] = UNSET
    project_id: Union[None, Unset, str] = UNSET
    run_tags: Union[Unset, list["RunTagCreateRequest"]] = UNSET
    winner: Union[None, Unset, bool] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        task_type = self.task_type

        created_by: Union[None, Unset, str]
        if isinstance(self.created_by, Unset):
            created_by = UNSET
        else:
            created_by = self.created_by

        dataset_hash: Union[None, Unset, str]
        if isinstance(self.dataset_hash, Unset):
            dataset_hash = UNSET
        else:
            dataset_hash = self.dataset_hash

        num_samples: Union[None, Unset, int]
        if isinstance(self.num_samples, Unset):
            num_samples = UNSET
        else:
            num_samples = self.num_samples

        project_id: Union[None, Unset, str]
        if isinstance(self.project_id, Unset):
            project_id = UNSET
        else:
            project_id = self.project_id

        run_tags: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.run_tags, Unset):
            run_tags = []
            for run_tags_item_data in self.run_tags:
                run_tags_item = run_tags_item_data.to_dict()
                run_tags.append(run_tags_item)

        winner: Union[None, Unset, bool]
        if isinstance(self.winner, Unset):
            winner = UNSET
        else:
            winner = self.winner

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"name": name, "task_type": task_type})
        if created_by is not UNSET:
            field_dict["created_by"] = created_by
        if dataset_hash is not UNSET:
            field_dict["dataset_hash"] = dataset_hash
        if num_samples is not UNSET:
            field_dict["num_samples"] = num_samples
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if run_tags is not UNSET:
            field_dict["run_tags"] = run_tags
        if winner is not UNSET:
            field_dict["winner"] = winner

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.run_tag_create_request import RunTagCreateRequest

        d = src_dict.copy()
        name = d.pop("name")

        task_type = d.pop("task_type")

        def _parse_created_by(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        created_by = _parse_created_by(d.pop("created_by", UNSET))

        def _parse_dataset_hash(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_hash = _parse_dataset_hash(d.pop("dataset_hash", UNSET))

        def _parse_num_samples(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        num_samples = _parse_num_samples(d.pop("num_samples", UNSET))

        def _parse_project_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        project_id = _parse_project_id(d.pop("project_id", UNSET))

        run_tags = []
        _run_tags = d.pop("run_tags", UNSET)
        for run_tags_item_data in _run_tags or []:
            run_tags_item = RunTagCreateRequest.from_dict(run_tags_item_data)

            run_tags.append(run_tags_item)

        def _parse_winner(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        winner = _parse_winner(d.pop("winner", UNSET))

        run_create = cls(
            name=name,
            task_type=task_type,
            created_by=created_by,
            dataset_hash=dataset_hash,
            num_samples=num_samples,
            project_id=project_id,
            run_tags=run_tags,
            winner=winner,
        )

        run_create.additional_properties = d
        return run_create

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.get_prompt_rows_response import GetPromptRowsResponse
from ...models.http_validation_error import HTTPValidationError
from ...models.task_type import TaskType
from ...types import UNSET, Response, Unset


def _get_kwargs(
    project_id: str, run_id: str, chain_id: str, *, task_type: Union[Unset, TaskType] = UNSET
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    json_task_type: Union[Unset, int] = UNSET
    if not isinstance(task_type, Unset):
        json_task_type = task_type.value

    params["task_type"] = json_task_type

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/projects/{project_id}/runs/{run_id}/chains/{chain_id}",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[GetPromptRowsResponse, HTTPValidationError]]:
    if response.status_code == 200:
        response_200 = GetPromptRowsResponse.from_dict(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[GetPromptRowsResponse, HTTPValidationError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str,
    run_id: str,
    chain_id: str,
    *,
    client: AuthenticatedClient,
    task_type: Union[Unset, TaskType] = UNSET,
) -> Response[Union[GetPromptRowsResponse, HTTPValidationError]]:
    """Get Prompt Chain Rows

    Args:
        project_id (str):
        run_id (str):
        chain_id (str):
        task_type (Union[Unset, TaskType]): Valid task types for modeling.

            We store these as ints instead of strings because we will be looking this up in the
            database frequently.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[GetPromptRowsResponse, HTTPValidationError]]
    """

    kwargs = _get_kwargs(project_id=project_id, run_id=run_id, chain_id=chain_id, task_type=task_type)

    response = client.get_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    project_id: str,
    run_id: str,
    chain_id: str,
    *,
    client: AuthenticatedClient,
    task_type: Union[Unset, TaskType] = UNSET,
) -> Optional[Union[GetPromptRowsResponse, HTTPValidationError]]:
    """Get Prompt Chain Rows

    Args:
        project_id (str):
        run_id (str):
        chain_id (str):
        task_type (Union[Unset, TaskType]): Valid task types for modeling.

            We store these as ints instead of strings because we will be looking this up in the
            database frequently.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[GetPromptRowsResponse, HTTPValidationError]
    """

    return sync_detailed(
        project_id=project_id, run_id=run_id, chain_id=chain_id, client=client, task_type=task_type
    ).parsed


async def asyncio_detailed(
    project_id: str,
    run_id: str,
    chain_id: str,
    *,
    client: AuthenticatedClient,
    task_type: Union[Unset, TaskType] = UNSET,
) -> Response[Union[GetPromptRowsResponse, HTTPValidationError]]:
    """Get Prompt Chain Rows

    Args:
        project_id (str):
        run_id (str):
        chain_id (str):
        task_type (Union[Unset, TaskType]): Valid task types for modeling.

            We store these as ints instead of strings because we will be looking this up in the
            database frequently.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[GetPromptRowsResponse, HTTPValidationError]]
    """

    kwargs = _get_kwargs(project_id=project_id, run_id=run_id, chain_id=chain_id, task_type=task_type)

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str,
    run_id: str,
    chain_id: str,
    *,
    client: AuthenticatedClient,
    task_type: Union[Unset, TaskType] = UNSET,
) -> Optional[Union[GetPromptRowsResponse, HTTPValidationError]]:
    """Get Prompt Chain Rows

    Args:
        project_id (str):
        run_id (str):
        chain_id (str):
        task_type (Union[Unset, TaskType]): Valid task types for modeling.

            We store these as ints instead of strings because we will be looking this up in the
            database frequently.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[GetPromptRowsResponse, HTTPValidationError]
    """

    return (
        await asyncio_detailed(
            project_id=project_id, run_id=run_id, chain_id=chain_id, client=client, task_type=task_type
        )
    ).parsed



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.bar_chart import BarChart


T = TypeVar("T", bound="MAPResponse")


@_attrs_define
class MAPResponse:
    """The main metric for Object Detection.

    mAP, or mean Average Precision, is an Object Detection metric that provides you with a general sense of the
    performance of your model.

    mAP is not simply the average precision for each class, it is much more technical
    https://towardsdatascience.com/breaking-down-mean-average-precision-map-ae462f623a52

    We also provide each error and its "score", which represents how much the model
    would have improved by fixing each value

        Attributes:
            ap_per_class (BarChart): A class to represent a basic bar chart.

                labels: List[str] the x axis labels
                values: List[int | float] the counts for each bar
            error_scores (BarChart): A class to represent a basic bar chart.

                labels: List[str] the x axis labels
                values: List[int | float] the counts for each bar
            map_ (float):
            map_threshold (float):
    """

    ap_per_class: "BarChart"
    error_scores: "BarChart"
    map_: float
    map_threshold: float
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        ap_per_class = self.ap_per_class.to_dict()

        error_scores = self.error_scores.to_dict()

        map_ = self.map_

        map_threshold = self.map_threshold

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {"ap_per_class": ap_per_class, "error_scores": error_scores, "map": map_, "map_threshold": map_threshold}
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.bar_chart import BarChart

        d = src_dict.copy()
        ap_per_class = BarChart.from_dict(d.pop("ap_per_class"))

        error_scores = BarChart.from_dict(d.pop("error_scores"))

        map_ = d.pop("map")

        map_threshold = d.pop("map_threshold")

        map_response = cls(ap_per_class=ap_per_class, error_scores=error_scores, map_=map_, map_threshold=map_threshold)

        map_response.additional_properties = d
        return map_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CartographCluster")


@_attrs_define
class CartographCluster:
    """Information about a Cartograph embedding cluster.

    Attributes:
        cluster_description (str):
        cluster_id (int):
        cluster_size (int):
        average_confidence (Union[None, Unset, float]):
        average_dep (Union[None, Unset, float]):
        average_f1 (Union[None, Unset, float]):
        cluster_summary (Union[None, Unset, str]):
        cluster_topic (Union[None, Unset, str]):
    """

    cluster_description: str
    cluster_id: int
    cluster_size: int
    average_confidence: Union[None, Unset, float] = UNSET
    average_dep: Union[None, Unset, float] = UNSET
    average_f1: Union[None, Unset, float] = UNSET
    cluster_summary: Union[None, Unset, str] = UNSET
    cluster_topic: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cluster_description = self.cluster_description

        cluster_id = self.cluster_id

        cluster_size = self.cluster_size

        average_confidence: Union[None, Unset, float]
        if isinstance(self.average_confidence, Unset):
            average_confidence = UNSET
        else:
            average_confidence = self.average_confidence

        average_dep: Union[None, Unset, float]
        if isinstance(self.average_dep, Unset):
            average_dep = UNSET
        else:
            average_dep = self.average_dep

        average_f1: Union[None, Unset, float]
        if isinstance(self.average_f1, Unset):
            average_f1 = UNSET
        else:
            average_f1 = self.average_f1

        cluster_summary: Union[None, Unset, str]
        if isinstance(self.cluster_summary, Unset):
            cluster_summary = UNSET
        else:
            cluster_summary = self.cluster_summary

        cluster_topic: Union[None, Unset, str]
        if isinstance(self.cluster_topic, Unset):
            cluster_topic = UNSET
        else:
            cluster_topic = self.cluster_topic

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {"cluster_description": cluster_description, "cluster_id": cluster_id, "cluster_size": cluster_size}
        )
        if average_confidence is not UNSET:
            field_dict["average_confidence"] = average_confidence
        if average_dep is not UNSET:
            field_dict["average_dep"] = average_dep
        if average_f1 is not UNSET:
            field_dict["average_f1"] = average_f1
        if cluster_summary is not UNSET:
            field_dict["cluster_summary"] = cluster_summary
        if cluster_topic is not UNSET:
            field_dict["cluster_topic"] = cluster_topic

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        cluster_description = d.pop("cluster_description")

        cluster_id = d.pop("cluster_id")

        cluster_size = d.pop("cluster_size")

        def _parse_average_confidence(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        average_confidence = _parse_average_confidence(d.pop("average_confidence", UNSET))

        def _parse_average_dep(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        average_dep = _parse_average_dep(d.pop("average_dep", UNSET))

        def _parse_average_f1(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        average_f1 = _parse_average_f1(d.pop("average_f1", UNSET))

        def _parse_cluster_summary(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        cluster_summary = _parse_cluster_summary(d.pop("cluster_summary", UNSET))

        def _parse_cluster_topic(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        cluster_topic = _parse_cluster_topic(d.pop("cluster_topic", UNSET))

        cartograph_cluster = cls(
            cluster_description=cluster_description,
            cluster_id=cluster_id,
            cluster_size=cluster_size,
            average_confidence=average_confidence,
            average_dep=average_dep,
            average_f1=average_f1,
            cluster_summary=cluster_summary,
            cluster_topic=cluster_topic,
        )

        cartograph_cluster.additional_properties = d
        return cartograph_cluster

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.data_row import DataRow
    from ..models.inference_model_metrics import InferenceModelMetrics


T = TypeVar("T", bound="InferenceResults")


@_attrs_define
class InferenceResults:
    """
    Attributes:
        has_next_page (bool):
        labels (list[str]):
        model_metrics (InferenceModelMetrics):
        sample_count (int):
        split_total_sample_count (int):
        data_metrics (Union[Unset, list['DataRow']]):
        span_count (Union[None, Unset, int]):
        split_total_span_count (Union[None, Unset, int]):
    """

    has_next_page: bool
    labels: list[str]
    model_metrics: "InferenceModelMetrics"
    sample_count: int
    split_total_sample_count: int
    data_metrics: Union[Unset, list["DataRow"]] = UNSET
    span_count: Union[None, Unset, int] = UNSET
    split_total_span_count: Union[None, Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        has_next_page = self.has_next_page

        labels = self.labels

        model_metrics = self.model_metrics.to_dict()

        sample_count = self.sample_count

        split_total_sample_count = self.split_total_sample_count

        data_metrics: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.data_metrics, Unset):
            data_metrics = []
            for data_metrics_item_data in self.data_metrics:
                data_metrics_item = data_metrics_item_data.to_dict()
                data_metrics.append(data_metrics_item)

        span_count: Union[None, Unset, int]
        if isinstance(self.span_count, Unset):
            span_count = UNSET
        else:
            span_count = self.span_count

        split_total_span_count: Union[None, Unset, int]
        if isinstance(self.split_total_span_count, Unset):
            split_total_span_count = UNSET
        else:
            split_total_span_count = self.split_total_span_count

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "has_next_page": has_next_page,
                "labels": labels,
                "model_metrics": model_metrics,
                "sample_count": sample_count,
                "split_total_sample_count": split_total_sample_count,
            }
        )
        if data_metrics is not UNSET:
            field_dict["data_metrics"] = data_metrics
        if span_count is not UNSET:
            field_dict["span_count"] = span_count
        if split_total_span_count is not UNSET:
            field_dict["split_total_span_count"] = split_total_span_count

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.data_row import DataRow
        from ..models.inference_model_metrics import InferenceModelMetrics

        d = src_dict.copy()
        has_next_page = d.pop("has_next_page")

        labels = cast(list[str], d.pop("labels"))

        model_metrics = InferenceModelMetrics.from_dict(d.pop("model_metrics"))

        sample_count = d.pop("sample_count")

        split_total_sample_count = d.pop("split_total_sample_count")

        data_metrics = []
        _data_metrics = d.pop("data_metrics", UNSET)
        for data_metrics_item_data in _data_metrics or []:
            data_metrics_item = DataRow.from_dict(data_metrics_item_data)

            data_metrics.append(data_metrics_item)

        def _parse_span_count(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        span_count = _parse_span_count(d.pop("span_count", UNSET))

        def _parse_split_total_span_count(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        split_total_span_count = _parse_split_total_span_count(d.pop("split_total_span_count", UNSET))

        inference_results = cls(
            has_next_page=has_next_page,
            labels=labels,
            model_metrics=model_metrics,
            sample_count=sample_count,
            split_total_sample_count=split_total_sample_count,
            data_metrics=data_metrics,
            span_count=span_count,
            split_total_span_count=split_total_span_count,
        )

        inference_results.additional_properties = d
        return inference_results

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.like_dislike_constraints import LikeDislikeConstraints
    from ..models.score_constraints import ScoreConstraints
    from ..models.star_constraints import StarConstraints
    from ..models.tags_constraints import TagsConstraints
    from ..models.text_constraints import TextConstraints


T = TypeVar("T", bound="FeedbackTemplateCreate")


@_attrs_define
class FeedbackTemplateCreate:
    """
    Attributes:
        constraints (Union['LikeDislikeConstraints', 'ScoreConstraints', 'StarConstraints', 'TagsConstraints',
            'TextConstraints']):
        name (str):
        criteria (Union[None, Unset, str]):
        include_explanation (Union[Unset, bool]):  Default: False.
    """

    constraints: Union[
        "LikeDislikeConstraints", "ScoreConstraints", "StarConstraints", "TagsConstraints", "TextConstraints"
    ]
    name: str
    criteria: Union[None, Unset, str] = UNSET
    include_explanation: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.like_dislike_constraints import LikeDislikeConstraints
        from ..models.score_constraints import ScoreConstraints
        from ..models.star_constraints import StarConstraints
        from ..models.tags_constraints import TagsConstraints

        constraints: dict[str, Any]
        if isinstance(self.constraints, LikeDislikeConstraints):
            constraints = self.constraints.to_dict()
        elif isinstance(self.constraints, StarConstraints):
            constraints = self.constraints.to_dict()
        elif isinstance(self.constraints, ScoreConstraints):
            constraints = self.constraints.to_dict()
        elif isinstance(self.constraints, TagsConstraints):
            constraints = self.constraints.to_dict()
        else:
            constraints = self.constraints.to_dict()

        name = self.name

        criteria: Union[None, Unset, str]
        if isinstance(self.criteria, Unset):
            criteria = UNSET
        else:
            criteria = self.criteria

        include_explanation = self.include_explanation

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"constraints": constraints, "name": name})
        if criteria is not UNSET:
            field_dict["criteria"] = criteria
        if include_explanation is not UNSET:
            field_dict["include_explanation"] = include_explanation

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.like_dislike_constraints import LikeDislikeConstraints
        from ..models.score_constraints import ScoreConstraints
        from ..models.star_constraints import StarConstraints
        from ..models.tags_constraints import TagsConstraints
        from ..models.text_constraints import TextConstraints

        d = src_dict.copy()

        def _parse_constraints(
            data: object,
        ) -> Union[
            "LikeDislikeConstraints", "ScoreConstraints", "StarConstraints", "TagsConstraints", "TextConstraints"
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                constraints_type_0 = LikeDislikeConstraints.from_dict(data)

                return constraints_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                constraints_type_1 = StarConstraints.from_dict(data)

                return constraints_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                constraints_type_2 = ScoreConstraints.from_dict(data)

                return constraints_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                constraints_type_3 = TagsConstraints.from_dict(data)

                return constraints_type_3
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            constraints_type_4 = TextConstraints.from_dict(data)

            return constraints_type_4

        constraints = _parse_constraints(d.pop("constraints"))

        name = d.pop("name")

        def _parse_criteria(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        criteria = _parse_criteria(d.pop("criteria", UNSET))

        include_explanation = d.pop("include_explanation", UNSET)

        feedback_template_create = cls(
            constraints=constraints, name=name, criteria=criteria, include_explanation=include_explanation
        )

        feedback_template_create.additional_properties = d
        return feedback_template_create

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="StageDB")


@_attrs_define
class StageDB:
    """
    Attributes:
        created_by (str):
        id (str):
        name (str): Name of the stage. Must be unique within the project.
        project_id (str): ID of the project to which this stage belongs.
        description (Union[None, Unset, str]): Optional human-readable description of the goals of this guardrail.
        paused (Union[Unset, bool]): Whether the action is enabled. If False, the action will not be applied. Default:
            False.
        type_ (Union[Unset, Any]): Type of the stage. Default: 'local'.
        version (Union[None, Unset, int]):
    """

    created_by: str
    id: str
    name: str
    project_id: str
    description: Union[None, Unset, str] = UNSET
    paused: Union[Unset, bool] = False
    type_: Union[Unset, Any] = "local"
    version: Union[None, Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_by = self.created_by

        id = self.id

        name = self.name

        project_id = self.project_id

        description: Union[None, Unset, str]
        if isinstance(self.description, Unset):
            description = UNSET
        else:
            description = self.description

        paused = self.paused

        type_ = self.type_

        version: Union[None, Unset, int]
        if isinstance(self.version, Unset):
            version = UNSET
        else:
            version = self.version

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"created_by": created_by, "id": id, "name": name, "project_id": project_id})
        if description is not UNSET:
            field_dict["description"] = description
        if paused is not UNSET:
            field_dict["paused"] = paused
        if type_ is not UNSET:
            field_dict["type"] = type_
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        created_by = d.pop("created_by")

        id = d.pop("id")

        name = d.pop("name")

        project_id = d.pop("project_id")

        def _parse_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        description = _parse_description(d.pop("description", UNSET))

        paused = d.pop("paused", UNSET)

        type_ = d.pop("type", UNSET)

        def _parse_version(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        version = _parse_version(d.pop("version", UNSET))

        stage_db = cls(
            created_by=created_by,
            id=id,
            name=name,
            project_id=project_id,
            description=description,
            paused=paused,
            type_=type_,
            version=version,
        )

        stage_db.additional_properties = d
        return stage_db

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="GroupByMetrics")


@_attrs_define
class GroupByMetrics:
    """Metrics for a cohort returned in columnar format.

    Each field in the class is of equal length (or len 0), mapping to the y-axis
    of each chart. The labels field contains the x-axis labels which apply to all
    y-axes for all charts created for this data.

    ex:
        labels: ["apple", "banana", "orange"]
        precision: [0.95, 0.55, 0.83]
        recall: [0.93, 0.25, 0.88]
        ...

        Attributes:
            confidence (Union[Unset, list[float]]):
            data_error_potential (Union[Unset, list[float]]):
            f1 (Union[Unset, list[float]]):
            ghost_span (Union[Unset, list[int]]):
            labels (Union[Unset, list[Union[bool, int, str]]]):
            missed_label (Union[Unset, list[int]]):
            precision (Union[Unset, list[float]]):
            recall (Union[Unset, list[float]]):
            span_shift (Union[Unset, list[int]]):
            support (Union[Unset, list[int]]):
            total_errors (Union[Unset, list[int]]):
            wrong_tag (Union[Unset, list[int]]):
    """

    confidence: Union[Unset, list[float]] = UNSET
    data_error_potential: Union[Unset, list[float]] = UNSET
    f1: Union[Unset, list[float]] = UNSET
    ghost_span: Union[Unset, list[int]] = UNSET
    labels: Union[Unset, list[Union[bool, int, str]]] = UNSET
    missed_label: Union[Unset, list[int]] = UNSET
    precision: Union[Unset, list[float]] = UNSET
    recall: Union[Unset, list[float]] = UNSET
    span_shift: Union[Unset, list[int]] = UNSET
    support: Union[Unset, list[int]] = UNSET
    total_errors: Union[Unset, list[int]] = UNSET
    wrong_tag: Union[Unset, list[int]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        confidence: Union[Unset, list[float]] = UNSET
        if not isinstance(self.confidence, Unset):
            confidence = self.confidence

        data_error_potential: Union[Unset, list[float]] = UNSET
        if not isinstance(self.data_error_potential, Unset):
            data_error_potential = self.data_error_potential

        f1: Union[Unset, list[float]] = UNSET
        if not isinstance(self.f1, Unset):
            f1 = self.f1

        ghost_span: Union[Unset, list[int]] = UNSET
        if not isinstance(self.ghost_span, Unset):
            ghost_span = self.ghost_span

        labels: Union[Unset, list[Union[bool, int, str]]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = []
            for labels_item_data in self.labels:
                labels_item: Union[bool, int, str]
                labels_item = labels_item_data
                labels.append(labels_item)

        missed_label: Union[Unset, list[int]] = UNSET
        if not isinstance(self.missed_label, Unset):
            missed_label = self.missed_label

        precision: Union[Unset, list[float]] = UNSET
        if not isinstance(self.precision, Unset):
            precision = self.precision

        recall: Union[Unset, list[float]] = UNSET
        if not isinstance(self.recall, Unset):
            recall = self.recall

        span_shift: Union[Unset, list[int]] = UNSET
        if not isinstance(self.span_shift, Unset):
            span_shift = self.span_shift

        support: Union[Unset, list[int]] = UNSET
        if not isinstance(self.support, Unset):
            support = self.support

        total_errors: Union[Unset, list[int]] = UNSET
        if not isinstance(self.total_errors, Unset):
            total_errors = self.total_errors

        wrong_tag: Union[Unset, list[int]] = UNSET
        if not isinstance(self.wrong_tag, Unset):
            wrong_tag = self.wrong_tag

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if confidence is not UNSET:
            field_dict["confidence"] = confidence
        if data_error_potential is not UNSET:
            field_dict["data_error_potential"] = data_error_potential
        if f1 is not UNSET:
            field_dict["f1"] = f1
        if ghost_span is not UNSET:
            field_dict["ghost_span"] = ghost_span
        if labels is not UNSET:
            field_dict["labels"] = labels
        if missed_label is not UNSET:
            field_dict["missed_label"] = missed_label
        if precision is not UNSET:
            field_dict["precision"] = precision
        if recall is not UNSET:
            field_dict["recall"] = recall
        if span_shift is not UNSET:
            field_dict["span_shift"] = span_shift
        if support is not UNSET:
            field_dict["support"] = support
        if total_errors is not UNSET:
            field_dict["total_errors"] = total_errors
        if wrong_tag is not UNSET:
            field_dict["wrong_tag"] = wrong_tag

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        confidence = cast(list[float], d.pop("confidence", UNSET))

        data_error_potential = cast(list[float], d.pop("data_error_potential", UNSET))

        f1 = cast(list[float], d.pop("f1", UNSET))

        ghost_span = cast(list[int], d.pop("ghost_span", UNSET))

        labels = []
        _labels = d.pop("labels", UNSET)
        for labels_item_data in _labels or []:

            def _parse_labels_item(data: object) -> Union[bool, int, str]:
                return cast(Union[bool, int, str], data)

            labels_item = _parse_labels_item(labels_item_data)

            labels.append(labels_item)

        missed_label = cast(list[int], d.pop("missed_label", UNSET))

        precision = cast(list[float], d.pop("precision", UNSET))

        recall = cast(list[float], d.pop("recall", UNSET))

        span_shift = cast(list[int], d.pop("span_shift", UNSET))

        support = cast(list[int], d.pop("support", UNSET))

        total_errors = cast(list[int], d.pop("total_errors", UNSET))

        wrong_tag = cast(list[int], d.pop("wrong_tag", UNSET))

        group_by_metrics = cls(
            confidence=confidence,
            data_error_potential=data_error_potential,
            f1=f1,
            ghost_span=ghost_span,
            labels=labels,
            missed_label=missed_label,
            precision=precision,
            recall=recall,
            span_shift=span_shift,
            support=support,
            total_errors=total_errors,
            wrong_tag=wrong_tag,
        )

        group_by_metrics.additional_properties = d
        return group_by_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.split import Split

if TYPE_CHECKING:
    from ..models.inference_results import InferenceResults


T = TypeVar("T", bound="InferenceRunResults")


@_attrs_define
class InferenceRunResults:
    """
    Attributes:
        inference_name (str):
        split_name (Split):
        split_run_results (InferenceResults):
    """

    inference_name: str
    split_name: Split
    split_run_results: "InferenceResults"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        inference_name = self.inference_name

        split_name = self.split_name.value

        split_run_results = self.split_run_results.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {"inference_name": inference_name, "split_name": split_name, "split_run_results": split_run_results}
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.inference_results import InferenceResults

        d = src_dict.copy()
        inference_name = d.pop("inference_name")

        split_name = Split(d.pop("split_name"))

        split_run_results = InferenceResults.from_dict(d.pop("split_run_results"))

        inference_run_results = cls(
            inference_name=inference_name, split_name=split_name, split_run_results=split_run_results
        )

        inference_run_results.additional_properties = d
        return inference_run_results

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
